import numpy as np
import inspect
import pyteos_air.teos_air as teos_air

def attribute_to_function(method_name,function_name,*args):
    #in_args=np.broadcast_arrays(*np.atleast_3d(*args))
    #in_args+=(np.empty_like(in_args[0]),)
    #getattr(teos_air,method_name)(function_name,*in_args)
    #return np.squeeze(in_args[-1])
    return np.squeeze(getattr(teos_air,method_name)(function_name,*np.broadcast_arrays(*np.atleast_3d(*args))))




realm=__name__.split('.')[-2]
input_type=__name__.split('.')[-1]

from ... import pyteos_interface
import sys, imp

docstring="""
:param A: dry air massfraction (kg/kg)
:type A: np.array.
:param T: absolute temperature (K)
:type T: np.array.
:param p: total pressure (Pa)
:type p: np.array.
:returns: {0}

This is for wet air with ice and liquid.
"""

def attfunc(function_name,desc):
    func=lambda A,T,p: pyteos_interface.attribute_to_function(realm+'_'+input_type,function_name,A,T,p)
    func.__name__=function_name
    func.__doc__=docstring.format(desc)
    return func

entropy=                attfunc("entropy","wet air entropy (J/K).")
cond_entropy=           attfunc("cond_entropy","wet air entropy with all the moisture in condensed phase (J/K)")
density=                attfunc("density","wet air density (kg/m3).")
enthalpy=               attfunc("enthalpy","wet air enthalpy (J).")
rh_wmo=                 attfunc("rh_wmo","relative humidity using WMO definition")
temperatureequi=        attfunc("temperatureequi","equivalent temperature (K)")
temperaturewetbulb=     attfunc("temperaturewetbulb","wet bulb temperature (K)")



realm=__name__.split('.')[-2]
input_type=__name__.split('.')[-1]

from ... import pyteos_interface
import sys, imp

docstring="""
:param A: dry air massfraction (kg/kg)
:type A: np.array.
:param eta: wet air entropy obtained from :func:`pyteos_air.liq_ice_air.g.entropy` (J/K)
:type eta: np.array.
:returns: {0}

This is for wet air with ice and liquid.
"""

def attfunc(function_name,desc):
    func=lambda A,eta: pyteos_interface.attribute_to_function(realm+'_'+input_type,function_name,A,eta)
    func.__name__=function_name
    func.__doc__=docstring.format(desc)
    return func

ifl=        attfunc("ifl","isentropic freezing level (Pa).\n\n\
For pressures higher than the **ifl**, there can be only liquid water.")
iml=        attfunc("iml","isentropic melting level (Pa).\n\n\
For pressures lower than the **iml**, there can be only ice.")



import numpy as np

def function_parameters(func):
    func_desc=getattr(func,'func_doc')
    return [ x.split(':')[1][6:] for x in func_desc.split('\n') if (len(x)>=6 and x[:6]==':param')]

def create_gridded_data(realm,input_type,func,thermo_axes,num_procs=1):
    params_list=function_parameters(getattr(getattr(realm,input_type),func))
    if input_type not in ['g','g_ref','h','sat']:
        raise NotImplementedError('Input type '+input_type+' is not implemented.')

    #Check that the number of function parameters equals the number of thermo_axes inputs:
    if len(params_list) != len(thermo_axes.keys()):
        raise NameError('class Interpolated data: the input thermodynamics axes must have the same number of axes as the functions inputs.')

    #Check that all the functions parameters are given:
    if 'A' in params_list:
            fixed_params_list=[ x if x is not 'A' else 'rh_wmo' for x in params_list]
    if not set(params_list).issubset(thermo_axes.keys()) and not set(fixed_params_list).issubset(thermo_axes.keys()):
        raise NameError('class Interpolated data: the input thermodynamics axes must contain all of the functions inputs.')

    if input_type=='g_ref' and not np.isscalar(thermo_axes['pref']):
        raise IndexError('class Interpolated data: if the function has more than three dimensions pref must be a scalar')

    #Start the multiprocessing pool:
    if num_procs>1:
        import multiprocessing as mproc
        pool=mproc.Pool(processes=int(num_procs))
    else:
        pool=None

    if input_type in ['g','g_ref','sat']:
        T =np.reshape(thermo_axes['T'],[1,len(thermo_axes['T']),1])
        p =np.reshape(thermo_axes['p'],[1,1,len(thermo_axes['p'])])

        if input_type in ['g','g_ref']:
            if 'rh_wmo' in thermo_axes.keys():
                rh_wmo = np.reshape(thermo_axes['rh_wmo'],[len(thermo_axes['rh_wmo']),1,1])
                a_sat=np.reshape(mp_vec_masked(getattr(getattr(realm,'sat'),'massfraction_air'),(T,p)),(1,len(thermo_axes['T']),len(thermo_axes['p'])))
                A = 1.0 / (1.0 + rh_wmo * (1.0 / a_sat - 1.0))
                #Fix problems close to saturation:
                A = np.where(rh_wmo>=1.0,A+1e-10,A)
            else:
                A = np.reshape(thermo_axes['A'],[len(thermo_axes['A']),1,1])
            #If temperature is less than 193.0, make it dry:
            A=np.where(T<193.0,1.0,A)

    elif input_type in ['h']:
        eta =np.reshape(thermo_axes['eta'],[1,len(thermo_axes['eta']),1])
        p =np.reshape(thermo_axes['p'],[1,1,len(thermo_axes['p'])])
        A = np.reshape(thermo_axes['A'],[len(thermo_axes['A']),1,1])
        
    if input_type=='g_ref':
        result=mp_vec_masked(getattr(getattr(realm,input_type),func),(A,T,p,thermo_axes['pref']),pool=pool)
    elif input_type=='g':
        result=mp_vec_masked(getattr(getattr(realm,input_type),func),(A,T,p),pool=pool)
    elif input_type=='sat':
        result=mp_vec_masked(getattr(getattr(realm,input_type),func),(T,p),pool=pool)
    elif input_type=='h':
        result=mp_vec_masked(getattr(getattr(realm,input_type),func),(A,eta,p),pool=pool)

    if num_procs>1:
        pool.close()

    return result

def mp_vec_masked(func,args,pool=None):
    fill_value=1e90
    #This function simplifies the use of the multiprocessing toolbox.
    if pool:
        args=np.broadcast_arrays(*args)
        num_procs=len(pool._pool)
        dims_frac_proc=np.ma.array(np.array(args[0].shape)/num_procs)
        dim_index=np.argmin(np.ma.masked_where(dims_frac_proc<1.0,dims_frac_proc))
        in_shape=list(args[0].shape)
        in_shape[dim_index]=1
        iter_list=[[func for x in range(0,args[0].shape[dim_index])]]
        for in_arr in args:
            iter_list.append(np.split(np.ma.filled(in_arr,fill_value),args[0].shape[dim_index],axis=dim_index))
        out_var = np.concatenate(map(lambda x: np.reshape(x,in_shape),
                                    pool.map(tuple_function,zip(*iter_list))),axis=dim_index)
    else:
        out_var = func(*args)
    return np.ma.masked_where(abs(out_var)>=fill_value,out_var)

def tuple_function(args):
    return args[0](*args[1:])

def pickle_function(realm,input_type,func,thermo_axes,file_to_pickle,num_procs=1):
    import pickle
    import pyteos_air.liq_ice_air as liq_ice_air
    func_instance=Interpolated_data(realm,input_type,func,thermo_axes,num_procs=num_procs)

    file_id=open(file_to_pickle,'w')
    pickle.dump(func_instance,file_id)
    file_id.close



import numpy as np
import scipy.interpolate as interp

from create_gridded_data import create_gridded_data

class Interpolated_data:
    def __init__(self,realm,input_type,func_name,thermo_axes,num_procs=1):
        #self._realm=realm
        self._input_type=input_type
        self.__name__=getattr(getattr(getattr(realm,input_type),func_name),'__name__')

        params_list=function_parameters(getattr(getattr(realm,input_type),func_name))
        if input_type in ['g','g_ref'] and 'rh_wmo' in thermo_axes.keys():
            params_list[0]='rh_wmo'
        docstring=self.__name__+'('+','.join(params_list)+')'+getattr(getattr(getattr(realm,input_type),func_name),'__doc__')
        docstring+='\nSpline interpolation from gridded data with grid:\n'

        thermo_data=create_gridded_data(realm,input_type,func_name,thermo_axes,num_procs=num_procs)

        if input_type=='sat':
            docstring+='T: '+str(thermo_axes['T'])+'\n'
            docstring+='p: '+str(thermo_axes['p'])+'\n'

            self._interpolants=interp_linear((thermo_axes['T'],thermo_axes['p'],np.squeeze(thermo_data)))
        elif input_type in ['h']:
            self._p = thermo_axes['p']
            num_p=len(self._p)

            docstring+='A: '+str(thermo_axes['A'])+'\n'
            docstring+='eta: '+str(thermo_axes['eta'])+'\n'
            docstring+='p: '+str(thermo_axes['p'])+'\n'
            docstring+='The interpolation is a spline on constant p and linear along p.'

            self._interpolants=interp_linear_3D((thermo_axes['A'],thermo_axes['eta'],thermo_axes['p'],thermo_data))
        elif input_type in ['g','g_ref']:
            self._p = thermo_axes['p']
            num_p=len(self._p)

            if 'rh_wmo' in thermo_axes.keys():
                moisture_var='rh_wmo'
                docstring=docstring.replace(':param A: dry air massfraction (kg/kg)',
                                            ':param rh_wmo: relative humidity, WMO definition, 0 < rh < 1 for unsaturated air')  
                docstring=docstring.replace(':type A:',
                                            ':type rh_wmo')
                docstring+='rh_wmo: '+str(thermo_axes['rh_wmo'])+'\n'
            else:
                moisture_var='A'
                docstring+='A: '+str(thermo_axes['A'])+'\n'

            docstring+='T: '+str(thermo_axes['T'])+'\n'
            docstring+='p: '+str(thermo_axes['p'])+'\n'
            docstring+='The interpolation is a spline on constant p and linear along p.'


            self._interpolants=interp_linear_3D((thermo_axes[moisture_var],thermo_axes['T'],thermo_axes['p'],thermo_data))
        self.__doc__=docstring

    def __call__(self,*args):
        if self._input_type=='sat':
            T, p = np.broadcast_arrays(*np.atleast_3d(*args))
            return np.reshape(self._interpolants.ev(np.ravel(T),np.ravel(p)),T.shape)
        elif self._input_type in ['h','g','g_ref']:
            return self._interpolants(np.concatenate([a[...,np.newaxis] for a in np.broadcast_arrays(*np.atleast_3d(*args))],axis=-1))
        else:
            return
        
def interp_linear(args):
    nd_interp=interp.RectBivariateSpline(*args,kx=1,ky=1)
    return nd_interp

def interp_linear_3D(args):
    nd_interp=interp.RegularGridInterpolator(args[:-1],args[-1],bounds_error=False)
    return nd_interp

def function_parameters(func):
    func_desc=getattr(func,'func_doc')
    return [ x.split(':')[1][6:] for x in func_desc.split('\n') if (len(x)>=6 and x[:6]==':param')]

def pickle_interpolants(realm,input_type,functions_list,thermo_axes,file_name,num_procs=1):
    """
    :param realm: Realm (e.g. 'lid_ice_air')
    :type realm: str.
    :param input_type: Type of arguments (e.g. 'g')
    :type T: str.
    :param functions_list: List of function (e.g. ['density','enthalpy'])
    :type functions_list: list.
    :param thermo_axes: dictionary of axes to interpolate to (e.g. {'A':[0.5,1.0],'T':[250,300],'p':[1e4,1e5]})
    :type thermo_axes: dict.
    :param file_name: Name of file to pickle to
    :type file_name: str.

    This function creates a pickle file that can be used with pyteos_nc. This greatly speeds up computation.
    """

    for function in functions:
        func_list[function]=Interpolated_data(liq_ice_air,realm,function,thermo_axes,num_procs=num_procs)
    file=open(file_name,'w')
    pickle.dump(func_list,file)




realm=__name__.split('.')[-2]
input_type=__name__.split('.')[-1]

from ... import pyteos_interface
import sys, imp

docstring="""
:param A: dry air massfraction (kg/kg)
:type A: np.array.
:param eta: wet air entropy obtained from :func:`pyteos_air.liq_ice_air.g.entropy` (J/K)
:type eta: np.array.
:param p: total pressure (Pa)
:type p: np.array.
:returns: {0}

This is for wet air with ice and liquid.
"""

def attfunc(function_name,desc):
    func=lambda A,eta,p: pyteos_interface.attribute_to_function(realm+'_'+input_type,function_name,A,eta,p)
    func.__name__=function_name
    func.__doc__=docstring.format(desc)
    return func

temperature=attfunc("temperature","absolute temperature (K).\n\n\
Note that this function is not an exact inverse of :func:`pyteos_air.liq_ice_air.g.entropy`. \
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

density=attfunc("density","density (kg/m3).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

enthalpy=attfunc("enthalpy","enthalpy (J).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

gc_gv=attfunc("gc_gv","gc_minus_gv (J).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

gd_gc=attfunc("gd_gc","gd_minus_gc (J).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

gd=attfunc("gd","gd (J).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

gc=attfunc("gc","gc (J).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")

gh_gc=attfunc("gh_gc","gh_minus_gc (J).\n\n\
At the freezing point, this function assumes that a combination of liquid water and ice will \
combine to make sure that the temperature is constant between the Isentropic Freezing level \
:func:`pyteos_air.liq_ice_air.il.ifl` and the Isentropic Melting Level :func:`pyteos_air.liq_ice_air.il.iml`")



# This Python file uses the following encoding: utf-8
import os

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

from setuptools import setup, find_packages

#BUILD TEOS-10. IT REQUIRES A SIGNATURE FILE THAT CAN BE GENERATED BY:
#f2py -m teos_air --overwrite-signature -h teos_air.pyf src/teos_air.F90

from numpy.distutils.core import setup, Extension
teos_files=[
'Constants_0.F90',
'Convert_0.F90',
'Maths_0.F90',
'Flu_1.F90',
'Ice_1.F90',
'Sal_1.F90',
'Air_1.F90',
'Flu_2.F90',
'Ice_2.F90',
'Sal_2.F90',
'Air_2.F90',
'Flu_3a.F90',
'Flu_3b.F90',
'Sea_3a.F90',
'Sea_3b.F90',
'Sea_3c.F90',
'Sea_3d.F90',
'Air_3a.F90',
'Air_3b.F90',
'Air_3c.F90',
'Liq_Vap_4.F90',
'Ice_Vap_4.F90',
'Ice_Liq_4.F90',
'Liq_Ice_Air_4.F90',
'Sea_Vap_4.F90',
'Sea_Liq_4.F90',
'Sea_Ice_4.F90',
'Sea_Ice_Vap_4.F90',
'Liq_Air_4a.F90',
'Sea_Air_4.F90',
'Liq_Air_4b.F90',
'Liq_Air_4c.F90',
'Liq_Air_4d.F90',
'Ice_Air_4a.F90',
'Ice_Air_4b.F90',
'Ice_Air_4c.F90',
'Ice_Air_4d.F90',
#'Liq_Ice_Air_5.F90',
'Liq_Ice_Air_5a.F90',
'Liq_Ice_Air_5b.F90',
'Liq_Ice_Air_5c.F90',
'Liq_Ice_Air_5d.F90',
#'Liq_Ice_Air_5e.F90',
'Flu_IF97_5.F90',
'Ice_Flu_5.F90',
'Liq_F03_5.F90',
'Sea_5a.F90',
'GSW_Library_5.F90',
'Convert_5.F90',
'Air_5.F90',
'OS2008_5.F90',
]
teos_files=['src/SIA_library/'+file for file in teos_files]

sign_file='src/teos_air.pyf'
sources_interface=['lib/liq_ice_air/liq_ice_air_f2py.F90']
teos_files.extend(sources_interface)

# generate signature file if it does not exist:
if not os.path.exists(sign_file):
    print ' '.join(sources_interface)
    os.system('f2py -m teos_air --overwrite-signature -h '+sign_file+' '+' '.join(sources_interface))

teos_files.append(sign_file)

package_name='pyteos_air'
packages=[package_name+'.'+name for name in find_packages('lib')]
packages.append(package_name)

setup(name=package_name,
    version = "0.91",
    packages=packages,
    package_dir = {package_name: 'lib'},

        # metadata for upload to PyPI
        author = "F. Laliberté, R. Feistel, D. G. Wright, D. R. Jackett, K. Miyagawa, J. H. Reissmann, W. Wagner, U. Overhoff, C. Guder, A. Feistel, G. M. Marion",
        author_email = "frederic.laliberte@utoronto.ca",
        description = "This package provides python bindings to the atmospheric component of TEOS-10. \
                       Also includes expanded metorological quantities like the equivalent potential temperature.",
        license = "BSD",
        keywords = "thermodynamics atmosphere climate",
        url = "www.teos-10.org",   # project home page, if any
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Science/Research",
            "Natural Language :: English",
            "License :: OSI Approved :: BSD License",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Fortran",
            "Topic :: Scientific/Engineering :: Atmospheric Science",
            "Topic :: Scientific/Engineering :: Mathematics"
        ],
        long_description=read('README'),
        install_requires = ['numpy>=1.6','scipy'],
        zip_safe=False,
        ext_package = 'pyteos_air.teos_air',
        ext_modules = [Extension('teos_air',
                                  teos_files
                                ),
                      ],
    )



realm=__name__.split('.')[-2]
input_type=__name__.split('.')[-1]

from ... import pyteos_interface
import sys, imp


docstring="""
:param T: absolute temperature (K)
:type T: np.array.
:param p: total pressure (Pa)
:type p: np.array.
:returns: {0}

This is for wet air with ice and liquid.
"""

def attfunc(function_name,desc):
    func=lambda T,p: pyteos_interface.attribute_to_function(realm+'_'+input_type,function_name,T,p)
    func.__name__=function_name
    func.__doc__=docstring.format(desc)
    return func

massfraction_air=attfunc("massfraction_air","saturation dry air massfraction (kg/kg).")



realm=__name__.split('.')[-2]
input_type=__name__.split('.')[-1]

from ... import pyteos_interface

docstring="""
:param A: dry air massfraction (kg/kg)
:type A: np.array.
:param T: absolute temperature (K)
:type T: np.array.
:param p: total pressure (Pa)
:type p: np.array.
:param pref: reference total pressure (Pa)
:type pref: np.array.
:returns: {0}

This is for wet air with ice and liquid.
"""

def attfunc(function_name,desc):
    func=lambda A,T,p,pref: pyteos_interface.attribute_to_function(realm+'_'+input_type,function_name,A,T,p,pref)
    func.__name__=function_name
    func.__doc__=docstring.format(desc)
    return func

#Functions list:
pottemp=                attfunc("pottemp","potential tempeature (K) at **pref**.")
pottempequisat=         attfunc("pottempequisat","saturation equivalent potential temperature (K) at **pref**.")
pottempequipseudo=      attfunc("pottempequipseudo","pseudo equivalent potential temperature (K) at **pref**.\n\n\
Integrates the pseudo adiabat until the temperature falls\
outside the range of validity of TEOS-10.")
pottempequi=            attfunc("pottempequi","equivalent potential tempeature (K) at **pref**.")

